/**
 * 
 * Copyright (c) 2020, Jose-Luis Lisani, joseluis.lisani@uib.es
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <vector>
#include <algorithm>

#include "library/io_png.h"
#include "library/io_RGB.h"
#include "library/utils.h"
#include "library/libftc.h"


//data structure to store the results of segmenting the Intensity histogram
struct Isegment {
    vector<int> pixels;
    double R, G, B; //average RGB values of pixels
    float Imin, Imax; //minimum and maximum intensity values of pixels
};

//data structure to store the results of segmenting the Saturation histogram
struct Ssegment {
    vector<int> pixels;
    vector<Isegment> Isegments;
    double R, G, B; //average RGB values of pixels
    float Smin, Smax; //minimum and maximum saturation values of pixels
};

//data structure to store the results of segmenting the Hue histogram
struct Hsegment {
    vector<int> pixels;
    vector<Ssegment> Ssegments;
    double R, G, B; //average RGB values of pixels
    float Hmin, Hmax; //minimum and maximum hue values of pixels
};

//data structure to store RGB values
struct dataRGB {
    unsigned char R, G, B;
};


//Compute gray palette
//Algorithm 6 in accompanying paper
vector<Isegment> compute_gray_palette(float *I, float *S, 
                                      unsigned char *R, 
                                      unsigned char *G, 
                                      unsigned char *B,
                                      float Smin, int nbinsI, 
                                      float qI, vector<int> ftcsegI, 
                                      int w, int h)
{
    vector<Isegment> Isegments;
    
    //assign same ID to all bins belonging to same segment
    int *idsegmentsI=new int[nbinsI];
    int nsegmentsI=ftcsegI.size()-1;
    for (int ii=0; ii < nsegmentsI; ii++) {
        for (int k=ftcsegI[ii]; k <= ftcsegI[ii+1]; k++) idsegmentsI[k]=ii;
    }

    //store information of each segment in a data structure (for each Hsegment)
    for (int ii=0; ii < nsegmentsI; ii++) {
        struct Isegment Iseg; //empty struct
        Iseg.R=0;
        Iseg.G=0;
        Iseg.B=0;
        Iseg.Imin=ftcsegI[ii]*qI;
        Iseg.Imax=ftcsegI[ii+1]*qI;
        Isegments.push_back(Iseg);
    }    

    //assign each pixel to one of the segments
    int idseg, iI;
    for (int n=0; n < w*h; n++) {
        if (S[n] <= Smin) {
            iI=(int) (I[n]/qI);
            if (iI >= nbinsI) iI=nbinsI-1;
            idseg=idsegmentsI[iI];
            Isegments[idseg].pixels.push_back(n);
            Isegments[idseg].R+=R[n];
            Isegments[idseg].G+=G[n];
            Isegments[idseg].B+=B[n];
        }
    }

    delete[] idsegmentsI;
    
    //Get average gray values for each segment
    for (int i=0; i < nsegmentsI; i++) {
        int Rmean, Gmean, Bmean, Imean;
        Rmean=(int) (Isegments[i].R/Isegments[i].pixels.size() + 0.5f);
        Gmean=(int) (Isegments[i].G/Isegments[i].pixels.size() + 0.5f);
        Bmean=(int) (Isegments[i].B/Isegments[i].pixels.size() + 0.5f);
        //average gray
        Imean=(int) ( ((float) Rmean + (float) Gmean + (float) Bmean)/3.0f + 0.5f) ;
        Isegments[i].R=Imean;
        Isegments[i].G=Imean;
        Isegments[i].B=Imean;
    }    

    return Isegments;
}


//Algorithm 7 in accompanying paper
vector<Hsegment> hue_segmentation(float *H, float *S, 
                                  unsigned char *R, 
                                  unsigned char *G, 
                                  unsigned char *B,
                                  float Smin, 
                                  int nbins, float qH, 
                                  vector<int> ftcseg,
                                  int w, int h,
                                  unsigned char optionGray)
{
    vector<Hsegment> Hsegments;

    int nsegments=(int) ftcseg.size();
    if ((nsegments == 2) && (ftcseg[0] == 0) && (ftcseg[1] == nbins-1)) {
        //only 1 segment
        nsegments=1;
    }
    
    //assign same ID to all bins belonging to same segment
    int *idsegments=new int[nbins];
    for (int i=0; i < (int) nsegments; i++) {
        if (i < (int) nsegments - 1) {
            for (int k=ftcseg[i]; k < ftcseg[i+1]; k++) idsegments[k]=i;
        } else {
            for (int k=ftcseg[i]; k < nbins; k++) idsegments[k]=i;
            for (int k=0; k < ftcseg[0]; k++) idsegments[k]=i;
        }
    }

    //store information of each segment in a data structure
    for (int i=0; i < nsegments; i++) {
        struct Hsegment Hseg; //empty struct
        Hseg.R=0;
        Hseg.G=0;
        Hseg.B=0;
        Hseg.Hmin=ftcseg[i]*qH;
        Hseg.Hmax=(i < (int) nsegments - 1)?(ftcseg[i+1]*qH):(ftcseg[0]*qH);
        Hsegments.push_back(Hseg);
    }    
    
    //assign each pixel to one of the segments
    int idseg, iH;
    for (int n=0; n < w*h; n++) {
        //if optionGray == 0 or Saturation is higher that Smin, assign pixel
        if ((S[n] > Smin) || (!optionGray)) { 
            iH=(int) (H[n]/qH);
            if (iH >= nbins) iH=nbins-1;
            idseg=idsegments[iH];
            Hsegments[idseg].pixels.push_back(n); //add pixel to Hsegment list 
            Hsegments[idseg].R+=R[n];
            Hsegments[idseg].G+=G[n];
            Hsegments[idseg].B+=B[n];
        }
    }    

    //Get average RGB values for each segment
    for (int i=0; i < nsegments; i++) {
        int Rmean, Gmean, Bmean;
        Rmean=(int) (Hsegments[i].R/Hsegments[i].pixels.size() + 0.5f);
        Gmean=(int) (Hsegments[i].G/Hsegments[i].pixels.size() + 0.5f);
        Bmean=(int) (Hsegments[i].B/Hsegments[i].pixels.size() + 0.5f);
        Hsegments[i].R=Rmean;
        Hsegments[i].G=Gmean;
        Hsegments[i].B=Bmean;
    }    
    
    delete[] idsegments;

    return Hsegments;
}


//Algorithm 8 in accompanying paper (case Saturation)
void channel_segmentation_saturation(float *S, 
                                     unsigned char *R, 
                                     unsigned char *G, 
                                     unsigned char *B,
                                     struct Hsegment &Hseg, int nbins, float q, 
                                     vector<int> ftcseg)
{
    int nsegments=ftcseg.size()-1;
    //assign same ID to all bins belonging to same segment
    int *idsegments=new int[nbins];
    for (int ii=0; ii < nsegments; ii++) {
        for (int k=ftcseg[ii]; k <= ftcseg[ii+1]; k++) idsegments[k]=ii;
    }

    //store information of each segment in a data structure (for each Hsegment)
    for (int ii=0; ii < nsegments; ii++) {
        struct Ssegment Sseg; //empty struct
        Sseg.R=0;
        Sseg.G=0;
        Sseg.B=0;
        Sseg.Smin=ftcseg[ii]*q;
        Sseg.Smax=ftcseg[ii+1]*q;
        Hseg.Ssegments.push_back(Sseg);
    }    

    //assign each pixel to one of the segments
    int idseg, i;
    for (int k=0; k < Hseg.pixels.size(); k++) {
        int n=Hseg.pixels[k];
        i=(int) (S[n]/q);
        if (i >= nbins) i=nbins-1;
        idseg=idsegments[i];
        Hseg.Ssegments[idseg].pixels.push_back(n);
        Hseg.Ssegments[idseg].R+=R[n];
        Hseg.Ssegments[idseg].G+=G[n];
        Hseg.Ssegments[idseg].B+=B[n];
    }

    //Get average RGB values for each segment
    for (int i=0; i < Hseg.Ssegments.size(); i++) {
        int Rmean, Gmean, Bmean;
        Rmean=(int) (Hseg.Ssegments[i].R/Hseg.Ssegments[i].pixels.size() + 0.5f);
        Gmean=(int) (Hseg.Ssegments[i].G/Hseg.Ssegments[i].pixels.size() + 0.5f);
        Bmean=(int) (Hseg.Ssegments[i].B/Hseg.Ssegments[i].pixels.size() + 0.5f);
        Hseg.Ssegments[i].R=Rmean;
        Hseg.Ssegments[i].G=Gmean;
        Hseg.Ssegments[i].B=Bmean;
    }    

    delete[] idsegments;

}



//Algorithm 8 in accompanying paper (case Intensity)
void channel_segmentation_intensity(float *I, 
                                    unsigned char *R, 
                                    unsigned char *G, 
                                    unsigned char *B,
                                    struct Ssegment &Sseg, int nbins, float q, 
                                    vector<int> ftcseg)                                  
{
    int nsegments=ftcseg.size()-1;
    //assign same ID to all bins belonging to same segment
    int *idsegments=new int[nbins];
    for (int ii=0; ii < nsegments; ii++) {
        for (int k=ftcseg[ii]; k <= ftcseg[ii+1]; k++) idsegments[k]=ii;
    }

    //store information of each segment in a data structure (for each Ssegment)
    for (int ii=0; ii < nsegments; ii++) {
        struct Isegment Iseg; //empty struct
        Iseg.R=0;
        Iseg.G=0;
        Iseg.B=0;
        Iseg.Imin=ftcseg[ii]*q;
        Iseg.Imax=ftcseg[ii+1]*q;
        Sseg.Isegments.push_back(Iseg);
    }    

    //assign each pixel to one of the segments
    int idseg, i;
    for (int k=0; k < Sseg.pixels.size(); k++) {
        int n=Sseg.pixels[k];
        i=(int) (I[n]/q);
        if (i >= nbins) i=nbins-1;
        idseg=idsegments[i];
        Sseg.Isegments[idseg].pixels.push_back(n);
        Sseg.Isegments[idseg].R+=R[n];
        Sseg.Isegments[idseg].G+=G[n];
        Sseg.Isegments[idseg].B+=B[n];
    }


    //Get average RGB values for each segment
    for (int i=0; i < Sseg.Isegments.size(); i++) {
        int Rmean, Gmean, Bmean;
        Rmean=(int) (Sseg.Isegments[i].R/Sseg.Isegments[i].pixels.size() + 0.5f);
        Gmean=(int) (Sseg.Isegments[i].G/Sseg.Isegments[i].pixels.size() + 0.5f);
        Bmean=(int) (Sseg.Isegments[i].B/Sseg.Isegments[i].pixels.size() + 0.5f);
        Sseg.Isegments[i].R=Rmean;
        Sseg.Isegments[i].G=Gmean;
        Sseg.Isegments[i].B=Bmean;
    }    

    delete[] idsegments;
}



//Compute color palette
//Algorithm 9 in accompanying paper
vector<Hsegment> compute_color_palette(float *H, float *S, float *I, 
                                       unsigned char *R, 
                                       unsigned char *G, 
                                       unsigned char *B,
                                       float Smin, 
                                       int nbins, int nbinsS, int nbinsI, 
                                       float qH, float qS, float qI, 
                                       vector<int> ftcseg,
                                       int w, int h, float eps,
                                       unsigned char optionGray)
{
    
    //Obtain list of pixels associated to each mode of the Hue histogram
    vector<Hsegment> Hsegments;
    Hsegments=hue_segmentation(H, S, R, G, B, Smin, nbins, qH, ftcseg, w, h, optionGray);
    int nsegmentsH=Hsegments.size();
        
    //Segment Saturation (for each Hsegment)
    int iS;
    float *histS=new float[nbinsS];
    for (int i=0; i < nsegmentsH; i++) {
        //create histogram of saturation values (only for pixels in the segment)
        memset(histS, 0, nbinsS*sizeof(float));
        for (int k=0; k < Hsegments[i].pixels.size(); k++) {
            iS=(int) (S[Hsegments[i].pixels[k]]/qS);
            if (iS >= nbinsS) iS=nbinsS-1;
            histS[iS]+=1.0f;
        }
        //segment saturation histogram
        vector<int> ftcsegS=FTCsegmentation(histS, nbinsS, eps, 0); //non-circular histogram
        
        //Obtain list of pixels associated to each mode of the Saturation histogram
        channel_segmentation_saturation(S, R, G, B, Hsegments[i], nbinsS, qS, ftcsegS);
        int nsegmentsS=Hsegments[i].Ssegments.size();
        
        //Segment Intensity (for each Ssegment in each Hsegment)
        int iI;
        float *histI=new float[nbinsI];
        for (int j=0; j < nsegmentsS; j++) {
            //create histogram of intensity values (only for pixels in the segment)
            memset(histI, 0, nbinsI*sizeof(float));
            for (int k=0; k < Hsegments[i].Ssegments[j].pixels.size(); k++) {
                iI=(int) (I[Hsegments[i].Ssegments[j].pixels[k]]/qI);
                if (iI >= nbinsI) iI=nbinsI-1;
                histI[iI]+=1.0f;            
            }
            //segment histogram
            vector<int> ftcsegI=FTCsegmentation(histI, nbinsI, eps, 0); //non-circular histogram
        
            //Obtain list of pixels associated to each mode of the Saturation histogram
            channel_segmentation_intensity(I, R, G, B, Hsegments[i].Ssegments[j], nbinsI, qI, ftcsegI);               
        }
        delete[] histI;

    }
    delete[] histS;
   
    return Hsegments;
}



void get_list_grays_colors(vector<Isegment> Isegments, 
                           vector<Hsegment> Hsegments, 
                           vector<dataRGB> &meanRGB_I, 
                           vector<dataRGB> &meanRGB_H, 
                           vector<dataRGB> &meanRGB_HS, 
                           vector<dataRGB> &meanRGB_HSI,
                           vector<int> &icolorsH, 
                           vector<int> &icolorsS)                                 
{

    //Get gray-levels
    for (int i=0; i < Isegments.size(); i++) {
        struct dataRGB rgb;
        rgb.R=(int) Isegments[i].R;
        rgb.G=(int) Isegments[i].G;
        rgb.B=(int) Isegments[i].B;
        meanRGB_I.push_back(rgb);
    }    

    printf("Result:\n");
    printf("Number of graylevels: %i\n", (int) meanRGB_I.size());
    
    //get average RGB value of Hue modes
    for (int i=0; i < Hsegments.size(); i++) {
        struct dataRGB rgb;
        rgb.R=(int) Hsegments[i].R;
        rgb.G=(int) Hsegments[i].G;
        rgb.B=(int) Hsegments[i].B;
        meanRGB_H.push_back(rgb);
    }    


    //get average RGB value of Hue-Saturation modes
    for (int i=0; i < Hsegments.size(); i++) {
        for (int j=0; j < Hsegments[i].Ssegments.size(); j++) {
            struct dataRGB rgb;
            rgb.R=(int) Hsegments[i].Ssegments[j].R;
            rgb.G=(int) Hsegments[i].Ssegments[j].G;
            rgb.B=(int) Hsegments[i].Ssegments[j].B;
            meanRGB_HS.push_back(rgb);
        }     
    }    
    
    //get average RGB value of Hue-Saturation-Intensity modes
    int ncolorsHSI=0;
    for (int i=0; i < Hsegments.size(); i++) {
        icolorsH.push_back(ncolorsHSI);
        for (int j=0; j < Hsegments[i].Ssegments.size(); j++) {
            icolorsS.push_back(ncolorsHSI);
            for (int k=0; k < Hsegments[i].Ssegments[j].Isegments.size(); k++) {
                struct dataRGB rgb;
                rgb.R=(int) Hsegments[i].Ssegments[j].Isegments[k].R;
                rgb.G=(int) Hsegments[i].Ssegments[j].Isegments[k].G;
                rgb.B=(int) Hsegments[i].Ssegments[j].Isegments[k].B;
                meanRGB_HSI.push_back(rgb);
                ncolorsHSI++;   
            }  
        }   
    }    

    printf("Number of colors: %i\n", (int) meanRGB_HSI.size());


}


void get_color_palette_image(vector<dataRGB> meanRGB_I, 
                             vector<dataRGB> meanRGB_H, 
                             vector<dataRGB> meanRGB_HS, 
                             vector<dataRGB> meanRGB_HSI,
                             vector<int> icolorsH,
                             vector<int> icolorsS,
                             unsigned char isGray,
                             unsigned char optionGray)
{
    int ncolorsI=meanRGB_I.size();
    int ncolorsH=meanRGB_H.size();
    int ncolorsHS=meanRGB_HS.size();
    int ncolorsHSI=meanRGB_HSI.size();
    
    int displaycolors=1;
    int displaygrays=(optionGray==1)?(1):(0);
    if (isGray) {
        ncolorsHSI=0;
        ncolorsHS=0;
        ncolorsH=0;
        displaycolors=0;
        displaygrays=1;
    }
    
    //create color palette image
    int wpatch=20; //color patches size
    int wmargin=5; //separation between patches
    int wpaletteC=(wpatch+wmargin)*ncolorsHSI+wmargin; //width of color palette
    int wpaletteG=(wpatch+wmargin)*ncolorsI+wmargin; //width of gray palette
    int wpalette; //total width of color palette
    int woffC, woffG; //offsets in position of color and gray palettes
    if (ncolorsHSI > ncolorsI) {
        wpalette=wpaletteC; 
        woffC=0;
        woffG=(wpaletteC-wpaletteG)/2;
    } else {
        wpalette=wpaletteG; 
        woffG=0;
        woffC=(wpaletteG-wpaletteC)/2;
    }
    //total height of color palette (3 rows of colors (if not gray image) + 1 row monochrome (if optionGray=1))
    int hpalette=(wpatch+wmargin)*(1*displaygrays+3*displaycolors)+wmargin;
    
    unsigned char *Rpal=new unsigned char[wpalette*hpalette];
    unsigned char *Gpal=new unsigned char[wpalette*hpalette];
    unsigned char *Bpal=new unsigned char[wpalette*hpalette];
    unsigned char *imgPal= (unsigned char*) malloc(3*wpalette*hpalette);
    //white background
    memset(Rpal, 255, wpalette*hpalette);
    memset(Gpal, 255, wpalette*hpalette);
    memset(Bpal, 255, wpalette*hpalette);
    
    int xshift, yshift, idcolor, x, xx, yy;
    //draw I colors
    yshift=wmargin;
    xshift=wmargin+woffG;
    idcolor=0;
    x=xshift;
    yy=0;
    xx=0;
    //first row
    while (idcolor < ncolorsI) {
        if (xx < wpatch) { //draw color
            Rpal[(xx+x)+(yy+yshift)*wpalette]=meanRGB_I[idcolor].R;
            Gpal[(xx+x)+(yy+yshift)*wpalette]=meanRGB_I[idcolor].G;
            Bpal[(xx+x)+(yy+yshift)*wpalette]=meanRGB_I[idcolor].B;
        }
        xx++;
        if (xx >= wpatch+wmargin) { //change color
            idcolor++;
            xx=0;
            x+=wpatch+wmargin;
        }
    } 
    //repeat image row wpatch times
    for (yy=1; yy < wpatch; yy++) {
        memcpy(Rpal+(yy+yshift)*wpalette, Rpal+(0+yshift)*wpalette, wpalette);
        memcpy(Gpal+(yy+yshift)*wpalette, Gpal+(0+yshift)*wpalette, wpalette);
        memcpy(Bpal+(yy+yshift)*wpalette, Bpal+(0+yshift)*wpalette, wpalette);
    }
    
    if (!isGray && optionGray) {
        //draw horizontal line separating grays from colors
        for (int x=0; x < wpalette; x++) {
            Rpal[x+(wmargin+wpatch+wmargin/2)*wpalette]=0;
            Gpal[x+(wmargin+wpatch+wmargin/2)*wpalette]=0;
            Bpal[x+(wmargin+wpatch+wmargin/2)*wpalette]=0;
        }
    }
    
    //draw HSI colors
    yshift=hpalette-(wpatch+wmargin);
    xshift=wmargin+woffC;
    idcolor=0;
    x=xshift;
    yy=0;
    xx=0;
    //first row
    while (idcolor < ncolorsHSI) {
        if (xx < wpatch) { //draw color
            Rpal[(xx+x)+(yy+yshift)*wpalette]=meanRGB_HSI[idcolor].R;
            Gpal[(xx+x)+(yy+yshift)*wpalette]=meanRGB_HSI[idcolor].G;
            Bpal[(xx+x)+(yy+yshift)*wpalette]=meanRGB_HSI[idcolor].B;
        }
        xx++;
        if (xx >= wpatch+wmargin) { //change color
            idcolor++;
            xx=0;
            x+=wpatch+wmargin;
        }
    }
    //repeat image row wpatch times
    for (yy=1; yy < wpatch; yy++) {
        memcpy(Rpal+(yy+yshift)*wpalette, Rpal+(0+yshift)*wpalette, wpalette);
        memcpy(Gpal+(yy+yshift)*wpalette, Gpal+(0+yshift)*wpalette, wpalette);
        memcpy(Bpal+(yy+yshift)*wpalette, Bpal+(0+yshift)*wpalette, wpalette);
    }
    
    //draw HS colors
    yshift-=(wpatch+wmargin);
    for (int idcolorS=0; idcolorS < ncolorsHS; idcolorS++) {
        //initial and final x coordinates of HSI patches associated to HS color
        int xfirst=wmargin+woffC+(wpatch+wmargin)*icolorsS[idcolorS];
        int xnext=(idcolorS < ncolorsHS-1)?(wmargin+woffC+(wpatch+wmargin)*icolorsS[idcolorS+1]):(wpalette);
        //display centered
        xshift=(xfirst+xnext)/2-wpatch/2-2;
        yy=0;
        //first row
        memset(Rpal+xshift+(yy+yshift)*wpalette, meanRGB_HS[idcolorS].R, wpatch);
        memset(Gpal+xshift+(yy+yshift)*wpalette, meanRGB_HS[idcolorS].G, wpatch);
        memset(Bpal+xshift+(yy+yshift)*wpalette, meanRGB_HS[idcolorS].B, wpatch);
        //copy row wpatch times
        for (yy=1; yy < wpatch; yy++) {
            memcpy(Rpal+(yy+yshift)*wpalette, Rpal+(0+yshift)*wpalette, wpalette);
            memcpy(Gpal+(yy+yshift)*wpalette, Gpal+(0+yshift)*wpalette, wpalette);
            memcpy(Bpal+(yy+yshift)*wpalette, Bpal+(0+yshift)*wpalette, wpalette);
        }      
        //draw vertical black line separating colors
        if (xnext < wpalette) {
            for (yy=wmargin+2*(wpatch+wmargin); yy < hpalette; yy++) {
                Rpal[xnext-wmargin/2+yy*wpalette]=0;
                Gpal[xnext-wmargin/2+yy*wpalette]=0;
                Bpal[xnext-wmargin/2+yy*wpalette]=0;
            }
        }
    }
    
    //draw H colors
    yshift-=(wpatch+wmargin);
    for (int idcolorH=0; idcolorH < ncolorsH; idcolorH++) {
        //initial and final x coordinates of HSI patches associated to H color
        int xfirst=wmargin+woffC+(wpatch+wmargin)*icolorsH[idcolorH];
        int xnext=(idcolorH < ncolorsH-1)?(wmargin+woffC+(wpatch+wmargin)*icolorsH[idcolorH+1]):(wpalette);
        //display centered
        xshift=(xfirst+xnext)/2-wpatch/2-2;
        //printf("Hcolor %i   xfirst=%i  xnext=%i   xshift=%i\n", idcolorH, xfirst, xnext, xshift);
        yy=0;
        //first row
        memset(Rpal+xshift+(yy+yshift)*wpalette, meanRGB_H[idcolorH].R, wpatch);
        memset(Gpal+xshift+(yy+yshift)*wpalette, meanRGB_H[idcolorH].G, wpatch);
        memset(Bpal+xshift+(yy+yshift)*wpalette, meanRGB_H[idcolorH].B, wpatch);
        //copy row wpatch times
        for (yy=1; yy < wpatch; yy++) {
            memcpy(Rpal+(yy+yshift)*wpalette, Rpal+(0+yshift)*wpalette, wpalette);
            memcpy(Gpal+(yy+yshift)*wpalette, Gpal+(0+yshift)*wpalette, wpalette);
            memcpy(Bpal+(yy+yshift)*wpalette, Bpal+(0+yshift)*wpalette, wpalette);
        }      
        //draw vertical black line separating colors
        if (xnext < wpalette) {
            for (yy=wmargin+(wpatch+wmargin); yy < hpalette; yy++) {
                Rpal[xnext-wmargin/2+yy*wpalette]=0;
                Gpal[xnext-wmargin/2+yy*wpalette]=0;
                Bpal[xnext-wmargin/2+yy*wpalette]=0;
            }
        }
    }
    
    
    
    //write output image
    RGB2output(Rpal, Gpal, Bpal, imgPal, wpalette*hpalette);
    io_png_write_u8("palette.png", imgPal, wpalette, hpalette, 3); //3 channels
    
    delete[] Rpal;
    delete[] Gpal;
    delete[] Bpal;
    free(imgPal);  

}


void get_segmented_image(vector<Isegment> Isegments, 
                         vector<Hsegment> Hsegments, 
                         unsigned char *R, unsigned char *G, unsigned char *B, 
                         int w, int h)
{

    //Get gray level segmented image 
    for (int i=0; i < Isegments.size(); i++) {
        int Rmean, Gmean, Bmean;
        Rmean=(int) Isegments[i].R;
        Gmean=(int) Isegments[i].G;
        Bmean=(int) Isegments[i].B;
        for (int n=0; n < Isegments[i].pixels.size(); n++) {
            R[Isegments[i].pixels[n]]=Rmean;
            G[Isegments[i].pixels[n]]=Gmean;
            B[Isegments[i].pixels[n]]=Bmean;
        }   
    }    

    //Get color segmented image
    for (int i=0; i < Hsegments.size(); i++) {
        for (int j=0; j < Hsegments[i].Ssegments.size(); j++) {
            for (int k=0; k < Hsegments[i].Ssegments[j].Isegments.size(); k++) {
                int Rmean, Gmean, Bmean;
                Rmean=(int) Hsegments[i].Ssegments[j].Isegments[k].R;
                Gmean=(int) Hsegments[i].Ssegments[j].Isegments[k].G;
                Bmean=(int) Hsegments[i].Ssegments[j].Isegments[k].B;
                for (int n=0; n < Hsegments[i].Ssegments[j].Isegments[k].pixels.size(); n++) {
                    R[Hsegments[i].Ssegments[j].Isegments[k].pixels[n]]=Rmean;
                    G[Hsegments[i].Ssegments[j].Isegments[k].pixels[n]]=Gmean;
                    B[Hsegments[i].Ssegments[j].Isegments[k].pixels[n]]=Bmean;
                }
            }
        }
    }
    

}



//Parameters:
//- input image
//- output image
//- option Gray: 1 for gray+color palette, 0 for only color palette
//- bin size for Hue histograms (default: 6 degrees)
//- bin size for Saturation histograms (default: 5)
//- bin size for Intensity histograms
//Output:
//- image displaying the computed color palette
//- original image with colors replaced by the colors in the palette
//
//This function implements Algorithm 11 in accompanying IPOL paper
int main(int argc, const char **argv)
{

    if (argc < 2) {
        printf("Usage: segmentIHSI input.png output.png [optionGray=1] [qH=6] [qS=5] [qI=5] [eps=1]\n");
        return EXIT_FAILURE;
    }

    const char *namein=argv[1];
    const char *nameout=argv[2];
    unsigned char optionGray=1;
    if (argc > 3) optionGray=(atoi(argv[3]) == 1);
    //size (in degrees) of histogram bins
    float qH=6;
    if (argc > 4) qH=atof(argv[4]);
    float qS=5;
    if (argc > 5) qS=atof(argv[5]);
    float qI=5;
    if (argc > 6) qI=atof(argv[6]);
    float eps=1;
    if (argc > 7) eps=atof(argv[7]);
    
    
    printf("Bin sizes: Hue=%i (degrees)  Saturation=%i  Intensity=%i    Eps: %f\n", 
            (int) qH, (int) qS, (int) qI, eps); 

    //Read input image
    size_t ww, hh;
    int w, h; //input image dimensions
    unsigned char *img; //image data
    unsigned char *R, *G, *B; //image data organized in R, G and B channels
    /* read the PNG input image */
    img = io_png_read_u8_rgb(namein, &ww, &hh);
    w=(int) ww;
    h=(int) hh;
    input2RGB(img, &R, &G, &B, w*h);

    //Convert RGN values to HSI
    float *H=new float[w*h];
    float *S=new float[w*h];
    float *I=new float[w*h];
    for (int n=0; n < w*h; n++)
        colorRGBtonewHSI((float) R[n], (float) G[n], (float) B[n], 
                         H[n], S[n], I[n]);
    
    //Compute number of bins of histograms
    //Hue histogram
    int nbinsH=(int) (360/qH);
    if (nbinsH*qH < 360) nbinsH++;

    int nbinsS=(int) (208/qS);
    if (nbinsS*qS < 208) nbinsS++;
    
    int nbinsI=(int) (256/qI);
    if (nbinsI*qI < 256) nbinsI++;
    
    //Compute minimum saturation value that prevents quantization problems 
    //in Hue histogram
    //Smin=minimum saturation that allows nbins different values in 
    //the discrete RGB cube
    float Smin=((float) nbinsH) / (2.0*M_PI);
    printf("Minimum saturation for hue and saturation computation=%2.2f\n", Smin);
    printf("(Colors with saturation below this value shall not be considered for the parsing of hue values)\n");
    
    //Create histogram of Hue values 
    //(only for pixels whose saturation is above Smin)
    float *histH=new float[nbinsH];
    memset(histH, 0, nbinsH*sizeof(float));
    int iH, nhighsat=0;
    //Create histogram of Intensity values 
    //(only for pixels whose saturation is below Smin)
    float *histI=new float[nbinsI];
    memset(histI, 0, nbinsI*sizeof(float));
    int iI;
    for (int n=0; n < w*h; n++) {
        if (S[n] > Smin) {
            iH=(int) (H[n]/qH);
            if (iH >= nbinsH) iH=nbinsH-1;
            histH[iH]+=1.0f;
            nhighsat++; //count number of pixels with high saturation
        } else {
            iI=(int) (I[n]/qI);
            if (iI >= nbinsI) iI=nbinsI-1;
            histI[iI]+=1.0f;         
        }
    }

    //boolean indicating whether the image is grayscale or color 
    unsigned char isGray=0; 
    if (nhighsat == 0) isGray=1;
    
    //Segment Intensity histogram (only if gray image, or optionGray active)
    vector<int> ftcsegI;
    if (isGray || optionGray) ftcsegI=FTCsegmentation(histI, nbinsI, eps, 0); //non-circular histogram
    delete[] histI; 

    //Segment Hue histogram (only if not gray image)
    vector<int> ftcsegH;
    if (!isGray) ftcsegH=FTCsegmentation(histH, nbinsH, eps, 1); //circular histogram
    delete[] histH;
    
    //Compute gray palette (only if gray image, or optionGray active)
    vector<Isegment> Isegments;
    if (isGray || optionGray) Isegments=compute_gray_palette(I, S, R, G, B, Smin, 
                                                             nbinsI, qI, ftcsegI, w, h);
    
    //Compute color palette (only if not gray image)
    vector<Hsegment> Hsegments;
    if (!isGray) Hsegments=compute_color_palette(H, S, I, R, G, B, Smin, 
                                                 nbinsH, nbinsS, nbinsI, 
                                                 qH, qS, qI, ftcsegH, w, h,
                                                 eps, optionGray);
                             
    //Get list of gray levels and colors
    vector<dataRGB> meanRGB_I,  meanRGB_H, meanRGB_HS, meanRGB_HSI;
    //auxiliary variables needed to build image of color palette
    vector<int> icolorsH; //index of each color in final palette associated to H colors
    vector<int> icolorsS; //index of each color in final palette associated to HS colors
    get_list_grays_colors(Isegments, Hsegments, 
                          meanRGB_I, meanRGB_H, 
                          meanRGB_HS, meanRGB_HSI,
                          icolorsH, icolorsS);                                  
                                      
    //Create color palette image 
    get_color_palette_image(meanRGB_I, meanRGB_H, meanRGB_HS, meanRGB_HSI, 
                            icolorsH, icolorsS, isGray, optionGray);
    
    //Create segmented image
    get_segmented_image(Isegments, Hsegments, R, G, B, w, h);
    
            
    //Write segmented image to file
    RGB2output(R, G, B, img, w*h);
    io_png_write_u8(nameout, img, w, h, 3); //3 channels


    free(img);
    delete[] R;
    delete[] G;
    delete[] B;
    delete[] H;
    delete[] S;
    delete[] I;
      
    return EXIT_SUCCESS;

}
